import rgbaCounter from '../counter/rgbaCounter';
const defaultOptions = {
    maxRuns: 10,
    withAlpha: false,
    randomSeed: false,
    nearestColors: true
};
/**
 * Extracts colors using the K-Means algorithm.
 * With that approach the colors can be found by finding the mean color on each group.
 * Where the colors are grouped based on their RGB distance.
 *
 * @param colors the list of all colors to be considered
 * @param count the number of colors to be extracted
 * @param options additional options {@see KMeansColorExtractorOptions} for more details
 */
const kMeansColorExtractor = (colors, count, options) => {
    const mergedOptions = Object.assign(Object.assign({}, defaultOptions), options);
    const { withAlpha, nearestColors } = mergedOptions;
    const uniqueColors = rgbaCounter(colors, withAlpha)
        .sort((a, b) => b.count - a.count)
        .map((it) => it.color);
    if (uniqueColors.length <= count)
        return uniqueColors;
    const foundColors = runKMeans(uniqueColors, count, mergedOptions);
    if (!nearestColors)
        return foundColors;
    return foundColors.map((color) => getNearestColor(uniqueColors, color, withAlpha));
};
export default kMeansColorExtractor;
const runKMeans = (colors, count, options) => {
    const { maxRuns, withAlpha, randomSeed } = options;
    let centroids = getInitialCentroids(colors, count, randomSeed);
    let clustering = getIndexedColors(colors, centroids, withAlpha);
    for (let i = 0; i < maxRuns; i++) {
        const newCentroids = recomputeCentroids(colors, clustering, count);
        const newClustering = getIndexedColors(colors, newCentroids, withAlpha);
        if (isEqualArray(clustering, newClustering)) {
            return newCentroids;
        }
        else {
            clustering = newClustering;
            centroids = newCentroids;
        }
    }
    return centroids;
};
/**
 * Creates an initial centroid list
 *
 * @param colors
 * @param count
 * @param randomSeed
 */
const getInitialCentroids = (colors, count, randomSeed) => {
    const emptyArray = new Array(count).fill(null);
    if (randomSeed) {
        const randomIndex = () => Math.floor(Math.random() * colors.length);
        return emptyArray.map(() => colors[randomIndex()]);
    }
    const chunkSize = Math.floor(colors.length / count);
    return emptyArray.map((_, index) => colors[index * chunkSize]);
};
/**
 * Index the colors based on the centroids.
 * Each color will be assigned to the centroid index that is closest to that color.
 *
 * @param colors
 * @param centroids
 * @param withAlpha
 */
const getIndexedColors = (colors, centroids, withAlpha) => {
    return colors.map((color) => getNearestColorIndex(centroids, color, withAlpha));
};
/**
 * Recalculate the median position/color for each centroid based on the assigned colors for each centroid.
 *
 * @param colors
 * @param clustering
 * @param count
 */
const recomputeCentroids = (colors, clustering, count) => {
    const centroids = new Array(count)
        .fill(null)
        .map(() => ({ sum: { r: 0, g: 0, b: 0, a: 0 }, count: 0 }));
    colors.forEach((color, index) => {
        const cluster = clustering[index];
        centroids[cluster].count++;
        centroids[cluster].sum.r += color.r;
        centroids[cluster].sum.g += color.g;
        centroids[cluster].sum.b += color.b;
        centroids[cluster].sum.a += color.a;
    });
    return centroids.map(({ sum, count }) => ({
        r: sum.r / count,
        g: sum.g / count,
        b: sum.b / count,
        a: sum.a / count
    }));
};
const getNearestColorIndex = (colors, color, withAlpha) => {
    const { index } = colors
        .reduce((acc, item, index) => {
        const distance = getColorDistance(color, item, withAlpha);
        return distance < acc.distance
            ? { index, distance }
            : acc;
    }, { index: Infinity, distance: Infinity });
    return index;
};
const getNearestColor = (colors, color, withAlpha) => {
    const index = getNearestColorIndex(colors, color, withAlpha);
    return colors[index];
};
/**
 * Calculates the Euclidean distance between two colors.
 *
 * @param a
 * @param b
 * @param withAlpha
 */
const getColorDistance = (a, b, withAlpha) => {
    const distance = Math.pow(a.r - b.r, 2) +
        Math.pow(a.g - b.g, 2) +
        Math.pow(a.b - b.b, 2);
    if (withAlpha)
        return distance + Math.pow(a.a - b.a, 2);
    return distance;
};
const isEqualArray = (a, b) => !!a.find((it, index) => it === b[index]);
//# sourceMappingURL=kMeansColorExtractor.js.map